@@ -2,7 +2,7 @@
  *
  * The Netwide Assembler is copyright (C) 1996 Simon Tatham and
  * Julian Hall. All rights reserved. The software is
- * redistributable under the licence given in the file "Licence"
+ * redistributable under the license given in the file "LICENSE"
  * distributed in the NASM archive.
  */
 
@@ -15,6 +15,7 @@
 #include <ctype.h>
 #include <inttypes.h>
 #include <limits.h>
+#include <time.h>
 
 #include "nasm.h"
 #include "nasmlib.h"
@@ -37,22 +38,24 @@ struct forwrefinfo {            /* info held on forward refs. */
 static int get_bits(char *value);
 static uint32_t get_cpu(char *cpu_str);
 static void parse_cmdline(int, char **);
-static void assemble_file(char *);
+static void assemble_file(char *, StrList **);
 static void register_output_formats(void);
 static void report_error_gnu(int severity, const char *fmt, ...);
 static void report_error_vc(int severity, const char *fmt, ...);
 static void report_error_common(int severity, const char *fmt,
                                 va_list args);
-static int is_suppressed_warning(int severity);
+static bool is_suppressed_warning(int severity);
 static void usage(void);
 static efunc report_error;
 
 static int using_debug_info, opt_verbose_info;
 bool tasm_compatible_mode = false;
-int pass0;
+int pass0, passn;
 int maxbits = 0;
 int globalrel = 0;
 
+time_t official_compile_time;
+
 static char inname[FILENAME_MAX];
 static char outname[FILENAME_MAX];
 static char listname[FILENAME_MAX];
@@ -85,24 +88,28 @@ enum op_type {
     op_normal,                  /* Preprocess and assemble */
     op_preprocess,              /* Preprocess only */
     op_depend,                  /* Generate dependencies */
-    op_depend_missing_ok,	/* Generate dependencies, missing OK */
 };
 static enum op_type operating_mode;
+/* Dependency flags */
+static bool depend_emit_phony = false;
+static bool depend_missing_ok = false;
+static const char *depend_target = NULL;
+static const char *depend_file = NULL;
 
 /*
  * Which of the suppressible warnings are suppressed. Entry zero
- * doesn't do anything. Initial defaults are given here.
+ * isn't an actual warning, but it used for -w+error/-Werror.
  */
-static bool suppressed[1 + ERR_WARN_MAX] = {
-    0, true, true, true, false, true, false, true, true, false
+static bool suppressed[ERR_WARN_MAX+1] = {
+    true, false, true, false, false, true, false, true, true, false
 };
 
 /*
  * The option names for the suppressible warnings. As before, entry
  * zero does nothing.
  */
-static const char *suppressed_names[1 + ERR_WARN_MAX] = {
-    NULL, "macro-params", "macro-selfref", "orphan-labels",
+static const char *suppressed_names[ERR_WARN_MAX+1] = {
+    "error", "macro-params", "macro-selfref", "orphan-labels",
     "number-overflow", "gnu-elf-extensions", "float-overflow",
     "float-denorm", "float-underflow", "float-toolong"
 };
@@ -111,12 +118,12 @@ static const char *suppressed_names[1 + ERR_WARN_MAX] = {
  * The explanations for the suppressible warnings. As before, entry
  * zero does nothing.
  */
-static const char *suppressed_what[1 + ERR_WARN_MAX] = {
-    NULL,
-    "macro calls with wrong no. of params",
-    "cyclic macro self-references",
+static const char *suppressed_what[ERR_WARN_MAX+1] = {
+    "treat warnings as errors",
+    "macro calls with wrong parameter count",
+    "cyclic macro references",
     "labels alone on lines without trailing `:'",
-    "numeric constants do not fit in 32 bits",
+    "numeric constants does not fit in 64 bits",
     "using 8- or 16-bit relocation in ELF32, a GNU extension",
     "floating point overflow",
     "floating point denormal",
@@ -130,7 +137,7 @@ static const char *suppressed_what[1 + ERR_WARN_MAX] = {
  * not preprocess their source file.
  */
 
-static void no_pp_reset(char *, int, efunc, evalfunc, ListGen *);
+static void no_pp_reset(char *, int, efunc, evalfunc, ListGen *, StrList **);
 static char *no_pp_getline(void);
 static void no_pp_cleanup(int);
 static Preproc no_pp = {
@@ -155,13 +162,134 @@ static void nasm_fputs(const char *line, FILE * outfile)
 {
     if (outfile) {
         fputs(line, outfile);
-        fputc('\n', outfile);
+        putc('\n', outfile);
     } else
         puts(line);
 }
 
+/* Convert a struct tm to a POSIX-style time constant */
+static int64_t posix_mktime(struct tm *tm)
+{
+    int64_t t;
+    int64_t y = tm->tm_year;
+
+    /* See IEEE 1003.1:2004, section 4.14 */
+
+    t = (y-70)*365 + (y-69)/4 - (y-1)/100 + (y+299)/400;
+    t += tm->tm_yday;
+    t *= 24;
+    t += tm->tm_hour;
+    t *= 60;
+    t += tm->tm_min;
+    t *= 60;
+    t += tm->tm_sec;
+
+    return t;
+}
+
+static void define_macros_early(void)
+{
+    char temp[128];
+    struct tm lt, *lt_p, gm, *gm_p;
+    int64_t posix_time;
+
+    lt_p = localtime(&official_compile_time);
+    if (lt_p) {
+	lt = *lt_p;
+
+	strftime(temp, sizeof temp, "__DATE__=\"%Y-%m-%d\"", &lt);
+	pp_pre_define(temp);
+	strftime(temp, sizeof temp, "__DATE_NUM__=%Y%m%d", &lt);
+	pp_pre_define(temp);
+	strftime(temp, sizeof temp, "__TIME__=\"%H:%M:%S\"", &lt);
+	pp_pre_define(temp);
+	strftime(temp, sizeof temp, "__TIME_NUM__=%H%M%S", &lt);
+	pp_pre_define(temp);
+    }
+
+    gm_p = gmtime(&official_compile_time);
+    if (gm_p) {
+	gm = *gm_p;
+
+	strftime(temp, sizeof temp, "__UTC_DATE__=\"%Y-%m-%d\"", &gm);
+	pp_pre_define(temp);
+	strftime(temp, sizeof temp, "__UTC_DATE_NUM__=%Y%m%d", &gm);
+	pp_pre_define(temp);
+	strftime(temp, sizeof temp, "__UTC_TIME__=\"%H:%M:%S\"", &gm);
+	pp_pre_define(temp);
+	strftime(temp, sizeof temp, "__UTC_TIME_NUM__=%H%M%S", &gm);
+	pp_pre_define(temp);
+    }
+
+    if (gm_p)
+	posix_time = posix_mktime(&gm);
+    else if (lt_p)
+	posix_time = posix_mktime(&lt);
+    else
+	posix_time = 0;
+
+    if (posix_time) {
+	snprintf(temp, sizeof temp, "__POSIX_TIME__=%"PRId64, posix_time);
+	pp_pre_define(temp);
+    }
+}
+
+static void define_macros_late(void)
+{
+    char temp[128];
+
+    snprintf(temp, sizeof(temp), "__OUTPUT_FORMAT__=%s\n",
+	     ofmt->shortname);
+    pp_pre_define(temp);
+}
+
+static void emit_dependencies(StrList *list)
+{
+    FILE *deps;
+    int linepos, len;
+    StrList *l, *nl;
+
+    if (depend_file && strcmp(depend_file, "-")) {
+	deps = fopen(depend_file, "w");
+	if (!deps) {
+	    report_error(ERR_NONFATAL|ERR_NOFILE|ERR_USAGE,
+			 "unable to write dependency file `%s'", depend_file);
+	    return;
+	}
+    } else {
+	deps = stdout;
+    }
+    
+    linepos = fprintf(deps, "%s:", depend_target);
+    for (l = list; l; l = l->next) {
+	len = strlen(l->str);
+	if (linepos + len > 62) {
+	    fprintf(deps, " \\\n ");
+	    linepos = 1;
+	}
+	fprintf(deps, " %s", l->str);
+	linepos += len+1;
+    }
+    fprintf(deps, "\n\n");
+    
+    for (l = list; l; l = nl) {
+	if (depend_emit_phony)
+	    fprintf(deps, "%s:\n\n", l->str);
+	    
+	nl = l->next;
+	nasm_free(l);
+    }
+
+    if (deps != stdout)
+	fclose(deps);
+}
+
 int main(int argc, char **argv)
 {
+    StrList *depend_list = NULL, **depend_ptr;
+
+    time(&official_compile_time);
+
     pass0 = 1;
     want_usage = terminate_after_phase = false;
     report_error = report_error_gnu;
@@ -179,6 +307,10 @@ int main(int argc, char **argv)
 
     register_output_formats();
 
+    /* Define some macros dependent on the runtime, but not
+       on the command line. */
+    define_macros_early();
+
     parse_cmdline(argc, argv);
 
     if (terminate_after_phase) {
@@ -197,29 +329,29 @@ int main(int argc, char **argv)
     eval_global_info(ofmt, lookup_label, &location);
 
     /* define some macros dependent of command-line */
-    {
-        char temp[64];
-        snprintf(temp, sizeof(temp), "__OUTPUT_FORMAT__=%s\n",
-                 ofmt->shortname);
-        pp_pre_define(temp);
-    }
+    define_macros_late();
+
+    depend_ptr = (depend_file || (operating_mode == op_depend))
+	? &depend_list : NULL;
+    if (!depend_target)
+	depend_target = outname;
 
     switch (operating_mode) {
-    case op_depend_missing_ok:
-	pp_include_path(NULL);	/* "assume generated" */
-	/* fall through */
     case op_depend:
         {
             char *line;
-            preproc->reset(inname, 0, report_error, evaluate, &nasmlist);
+
+	    if (depend_missing_ok)
+		pp_include_path(NULL);	/* "assume generated" */
+
+            preproc->reset(inname, 0, report_error, evaluate, &nasmlist,
+			   depend_ptr);
             if (outname[0] == '\0')
                 ofmt->filename(inname, outname, report_error);
             ofile = NULL;
-            fprintf(stdout, "%s: %s", outname, inname);
             while ((line = preproc->getline()))
                 nasm_free(line);
             preproc->cleanup(0);
-            putc('\n', stdout);
         }
         break;
 
@@ -241,8 +373,10 @@ int main(int argc, char **argv)
 
             location.known = false;
 
-/*      pass = 1; */
-            preproc->reset(inname, 2, report_error, evaluate, &nasmlist);
+	    /* pass = 1; */
+            preproc->reset(inname, 2, report_error, evaluate, &nasmlist,
+			   depend_ptr);
+
             while ((line = preproc->getline())) {
                 /*
                  * We generate %line directives if needed for later programs
@@ -298,7 +432,7 @@ int main(int argc, char **argv)
 
             ofmt->init(ofile, report_error, define_label, evaluate);
 
-            assemble_file(inname);
+            assemble_file(inname, depend_ptr);
 
             if (!terminate_after_phase) {
                 ofmt->cleanup(using_debug_info);
@@ -318,6 +452,9 @@ int main(int argc, char **argv)
         break;
     }
 
+    if (depend_list)
+	emit_dependencies(depend_list);
+
     if (want_usage)
         usage();
 
@@ -336,9 +473,9 @@ int main(int argc, char **argv)
  * Get a parameter for a command line option.
  * First arg must be in the form of e.g. -f...
  */
-static char *get_param(char *p, char *q, int *advance)
+static char *get_param(char *p, char *q, bool *advance)
 {
-    *advance = 0;
+    *advance = false;
     if (p[2]) {                 /* the parameter's in the option */
         p += 2;
         while (isspace(*p))
@@ -346,7 +483,7 @@ static char *get_param(char *p, char *q, int *advance)
         return p;
     }
     if (q && q[0]) {
-        *advance = 1;
+        *advance = true;
         return q;
     }
     report_error(ERR_NONFATAL | ERR_NOFILE | ERR_USAGE,
@@ -354,6 +491,102 @@ static char *get_param(char *p, char *q, int *advance)
     return NULL;
 }
 
+/*
+ * Copy a filename
+ */
+static void copy_filename(char *dst, const char *src)
+{
+    size_t len = strlen(src);
+
+    if (len >= (size_t)FILENAME_MAX) {
+	report_error(ERR_FATAL | ERR_NOFILE, "file name too long");
+	return;
+    }
+    strncpy(dst, src, FILENAME_MAX);
+}
+
+/*
+ * Convert a string to Make-safe form
+ */
+static char *quote_for_make(const char *str)
+{
+    const char *p;
+    char *os, *q;
+
+    size_t n = 1;		/* Terminating zero */
+    size_t nbs = 0;
+
+    if (!str)
+	return NULL;
+
+    for (p = str; *p; p++) {
+	switch (*p) {
+	case ' ':
+	case '\t':
+	    /* Convert N backslashes + ws -> 2N+1 backslashes + ws */
+	    n += nbs + 2;
+	    nbs = 0;
+	    break;
+	case '$':
+	case '#':
+	    nbs = 0;
+	    n += 2;
+	    break;
+	case '\\':
+	    nbs++;
+	    n++;
+	    break;
+	default:
+	    nbs = 0;
+	    n++;
+	break;
+	}
+    }
+
+    /* Convert N backslashes at the end of filename to 2N backslashes */
+    if (nbs)
+	n += nbs;
+
+    os = q = nasm_malloc(n);
+
+    nbs = 0;
+    for (p = str; *p; p++) {
+	switch (*p) {
+	case ' ':
+	case '\t':
+	    while (nbs--)
+		*q++ = '\\';
+	    *q++ = '\\';
+	    *q++ = *p;
+	    break;
+	case '$':
+	    *q++ = *p;
+	    *q++ = *p;
+	    nbs = 0;
+	    break;
+	case '#':
+	    *q++ = '\\';
+	    *q++ = *p;
+	    nbs = 0;
+	    break;
+	case '\\':
+	    *q++ = *p;
+	    nbs++;
+	    break;
+	default:
+	    *q++ = *p;
+	    nbs = 0;
+	break;
+	}
+    }
+    while (nbs--)
+	*q++ = '\\';
+
+    *q = '\0';
+
+    return os;
+}
+
 struct textargs {
     const char *label;
     int value;
@@ -367,125 +600,141 @@ struct textargs textopts[] = {
     {NULL, 0}
 };
 
-int stopoptions = 0;
-static int process_arg(char *p, char *q)
+static bool stopoptions = false;
+static bool process_arg(char *p, char *q)
 {
     char *param;
-    int i, advance = 0;
+    int i;
+    bool advance = false;
+    bool suppress;
 
     if (!p || !p[0])
-        return 0;
+        return false;
 
     if (p[0] == '-' && !stopoptions) {
+	if (strchr("oOfpPdDiIlFXuUZwW", p[1])) {
+	    /* These parameters take values */
+	    if (!(param = get_param(p, q, &advance)))
+		return advance;
+	}
+
         switch (p[1]) {
         case 's':
             error_file = stdout;
             break;
-        case 'o':              /* these parameters take values */
-        case 'O':
-        case 'f':
-        case 'p':
-        case 'P':
-        case 'd':
-        case 'D':
-        case 'i':
-        case 'I':
-        case 'l':
-        case 'F':
-        case 'X':
-        case 'u':
-        case 'U':
-	case 'Z':
-            if (!(param = get_param(p, q, &advance)))
-                break;
-            if (p[1] == 'o') {  /* output file */
-                strcpy(outname, param);
-            } else if (p[1] == 'f') {   /* output format */
-                ofmt = ofmt_find(param);
-                if (!ofmt) {
-                    report_error(ERR_FATAL | ERR_NOFILE | ERR_USAGE,
-                                 "unrecognised output format `%s' - "
-                                 "use -hf for a list", param);
-                } else
-                    ofmt->current_dfmt = ofmt->debug_formats[0];
-            } else if (p[1] == 'O') {   /* Optimization level */
-                int opt;
 
-		if (!*param) {
-		    /* Naked -O == -Ox */
-		    optimizing = INT_MAX >> 1; /* Almost unlimited */
-		} else {
-		    while (*param) {
-			switch (*param) {
-			case '0': case '1': case '2': case '3': case '4':
-			case '5': case '6': case '7': case '8': case '9':
-			    opt = strtoul(param, &param, 10);
+	case 'o':		/* output file */
+	    copy_filename(outname, param);
+	    break;
 
-			    /* -O0 -> optimizing == -1, 0.98 behaviour */
-			    /* -O1 -> optimizing == 0, 0.98.09 behaviour */
-			    if (opt < 2)
-				optimizing = opt - 1;
-			    else if (opt <= 5)
-				/* The optimizer seems to have problems with
-				   < 5 passes?  Hidden bug? */
-				optimizing = 5;	/* 5 passes */
-			    else
-				optimizing = opt;   /* More than 5 passes */
-			    break;
+	case 'f':		/* output format */
+	    ofmt = ofmt_find(param);
+	    if (!ofmt) {
+		report_error(ERR_FATAL | ERR_NOFILE | ERR_USAGE,
+			     "unrecognised output format `%s' - "
+			     "use -hf for a list", param);
+	    } else {
+		ofmt->current_dfmt = ofmt->debug_formats[0];
+	    }
+	    break;
 
-			case 'v':
-			case '+':
-			    param++;
-			    opt_verbose_info = true;
-			    break;
+	case 'O':		/* Optimization level */
+	{
+	    int opt;
 
-			case 'x':
-			    param++;
-			    optimizing = INT_MAX >> 1; /* Almost unlimited */
-			    break;
+	    if (!*param) {
+		/* Naked -O == -Ox */
+		optimizing = INT_MAX >> 1; /* Almost unlimited */
+	    } else {
+		while (*param) {
+		    switch (*param) {
+		    case '0': case '1': case '2': case '3': case '4':
+		    case '5': case '6': case '7': case '8': case '9':
+			opt = strtoul(param, &param, 10);
 
-			default:
-			    report_error(ERR_FATAL,
-					 "unknown optimization option -O%c\n",
-					 *param);
-			    break;
-			}
+			/* -O0 -> optimizing == -1, 0.98 behaviour */
+			/* -O1 -> optimizing == 0, 0.98.09 behaviour */
+			if (opt < 2)
+			    optimizing = opt - 1;
+			else
+			    optimizing = opt;
+			break;
+
+		    case 'v':
+		    case '+':
+			param++;
+			opt_verbose_info = true;
+			break;
+
+		    case 'x':
+			param++;
+			optimizing = INT_MAX >> 1; /* Almost unlimited */
+			break;
+
+		    default:
+			report_error(ERR_FATAL,
+				     "unknown optimization option -O%c\n",
+				     *param);
+			break;
 		    }
 		}
-            } else if (p[1] == 'P' || p[1] == 'p') {    /* pre-include */
-                pp_pre_include(param);
-            } else if (p[1] == 'D' || p[1] == 'd') {    /* pre-define */
-                pp_pre_define(param);
-            } else if (p[1] == 'U' || p[1] == 'u') {    /* un-define */
-                pp_pre_undefine(param);
-            } else if (p[1] == 'I' || p[1] == 'i') {    /* include search path */
-                pp_include_path(param);
-            } else if (p[1] == 'l') {   /* listing file */
-                strcpy(listname, param);
-            } else if (p[1] == 'Z') {   /* error messages file */
-                strcpy(errname, param);
-            } else if (p[1] == 'F') {   /* specify debug format */
-                ofmt->current_dfmt = dfmt_find(ofmt, param);
-                if (!ofmt->current_dfmt) {
-                    report_error(ERR_FATAL | ERR_NOFILE | ERR_USAGE,
-                                 "unrecognized debug format `%s' for"
-                                 " output format `%s'",
-                                 param, ofmt->shortname);
-                }
-            } else if (p[1] == 'X') {   /* specify error reporting format */
-                if (nasm_stricmp("vc", param) == 0)
-                    report_error = report_error_vc;
-                else if (nasm_stricmp("gnu", param) == 0)
-                    report_error = report_error_gnu;
-                else
-                    report_error(ERR_FATAL | ERR_NOFILE | ERR_USAGE,
-                                 "unrecognized error reporting format `%s'",
-                                 param);
-            }
+	    }
+	    break;
+	}
+
+	case 'p':			/* pre-include */
+	case 'P':
+	    pp_pre_include(param);
+	    break;
+
+	case 'd':			/* pre-define */
+	case 'D':
+	    pp_pre_define(param);
+	    break;
+
+	case 'u':			/* un-define */
+	case 'U':
+	    pp_pre_undefine(param);
+	    break;
+
+	case 'i':			/* include search path */
+	case 'I':
+	    pp_include_path(param);
+	    break;
+
+	case 'l':			/* listing file */
+	    copy_filename(listname, param);
+	    break;
+
+	case 'Z':			/* error messages file */
+	    strcpy(errname, param);
+	    break;
+
+	case 'F':			/* specify debug format */
+	    ofmt->current_dfmt = dfmt_find(ofmt, param);
+	    if (!ofmt->current_dfmt) {
+		report_error(ERR_FATAL | ERR_NOFILE | ERR_USAGE,
+			     "unrecognized debug format `%s' for"
+			     " output format `%s'",
+			     param, ofmt->shortname);
+	    }
+	    break;
+
+	case 'X':		/* specify error reporting format */
+	    if (nasm_stricmp("vc", param) == 0)
+		report_error = report_error_vc;
+	    else if (nasm_stricmp("gnu", param) == 0)
+		report_error = report_error_gnu;
+	    else
+		report_error(ERR_FATAL | ERR_NOFILE | ERR_USAGE,
+			     "unrecognized error reporting format `%s'",
+			     param);
             break;
+
         case 'g':
             using_debug_info = true;
             break;
+
         case 'h':
             printf
                 ("usage: nasm [-@ response file] [-o outfile] [-f format] "
@@ -509,9 +758,10 @@ static int process_arg(char *p, char *q)
                  "    -D<macro>[=<value>] pre-defines a macro\n"
                  "    -U<macro>   undefines a macro\n"
                  "    -X<format>  specifies error reporting format (gnu or vc)\n"
-                 "    -w+foo      enables warnings about foo; -w-foo disables them\n"
-                 "where foo can be:\n");
-            for (i = 1; i <= ERR_WARN_MAX; i++)
+                 "    -w+foo      enables warning foo (equiv. -Wfoo)\n"
+		 "    -w-foo      disable warning foo (equiv. -Wno-foo)\n"
+                 "Warnings:\n");
+            for (i = 0; i <= ERR_WARN_MAX; i++)
                 printf("    %-23s %s (default %s)\n",
                        suppressed_names[i], suppressed_what[i],
                        suppressed[i] ? "off" : "on");
@@ -528,15 +778,18 @@ static int process_arg(char *p, char *q)
             }
             exit(0);            /* never need usage message here */
             break;
+
         case 'y':
             printf("\nvalid debug formats for '%s' output format are"
                    " ('*' denotes default):\n", ofmt->shortname);
             dfmt_list(ofmt, stdout);
             exit(0);
             break;
+
         case 't':
             tasm_compatible_mode = true;
             break;
+
         case 'v':
             {
                 const char *nasm_version_string =
@@ -549,30 +802,85 @@ static int process_arg(char *p, char *q)
                 exit(0);        /* never need usage message here */
             }
             break;
+
         case 'e':              /* preprocess only */
 	case 'E':
             operating_mode = op_preprocess;
             break;
+
         case 'a':              /* assemble only - don't preprocess */
             preproc = &no_pp;
             break;
+
+	case 'W':
+	    if (param[0] == 'n' && param[1] == 'o' && param[2] == '-') {
+		suppress = true;
+		param += 3;
+	    } else {
+		suppress = false;
+	    }
+	    goto set_warning;
+
         case 'w':
-            if (p[2] != '+' && p[2] != '-') {
+            if (param[0] != '+' && param[0] != '-') {
                 report_error(ERR_NONFATAL | ERR_NOFILE | ERR_USAGE,
                              "invalid option to `-w'");
-            } else {
-                for (i = 1; i <= ERR_WARN_MAX; i++)
-                    if (!nasm_stricmp(p + 3, suppressed_names[i]))
-                        break;
-                if (i <= ERR_WARN_MAX)
-                    suppressed[i] = (p[2] == '-');
-                else
-                    report_error(ERR_NONFATAL | ERR_NOFILE | ERR_USAGE,
-                                 "invalid option to `-w'");
+		break;
             }
+	    suppress = (param[0] == '-');
+	    param++;
+	    goto set_warning;
+	set_warning:
+	    for (i = 0; i <= ERR_WARN_MAX; i++)
+		if (!nasm_stricmp(param, suppressed_names[i]))
+		    break;
+	    if (i <= ERR_WARN_MAX)
+		suppressed[i] = suppress;
+	    else if (!nasm_stricmp(param, "all"))
+		for (i = 1; i <= ERR_WARN_MAX; i++)
+		    suppressed[i] = suppress;
+	    else if (!nasm_stricmp(param, "none"))
+		for (i = 1; i <= ERR_WARN_MAX; i++)
+		    suppressed[i] = !suppress;
+	    else
+		report_error(ERR_NONFATAL | ERR_NOFILE | ERR_USAGE,
+			     "invalid warning `%s'", param);
             break;
+
         case 'M':
-            operating_mode = p[2] == 'G' ? op_depend_missing_ok : op_depend;
+	    switch (p[2]) {
+	    case 0:
+		operating_mode = op_depend;
+		break;
+	    case 'G':
+		operating_mode = op_depend;
+		depend_missing_ok = true;
+		break;
+	    case 'P':
+		depend_emit_phony = true;
+		break;
+	    case 'D':
+		depend_file = q;
+		advance = true;
+		break;
+	    case 'T':
+		depend_target = q;
+		advance = true;
+		break;
+	    case 'Q':
+		depend_target = quote_for_make(q);
+		advance = true;
+		break;
+	    default:
+		report_error(ERR_NONFATAL|ERR_NOFILE|ERR_USAGE,
+			     "unknown dependency option `-M%c'", p[2]);
+		break;
+	    }
+	    if (advance && (!q || !q[0])) {
+		report_error(ERR_NONFATAL|ERR_NOFILE|ERR_USAGE,
+			     "option `-M%c' requires a parameter", p[2]);
+		break;
+	    }
             break;
 
         case '-':
@@ -636,8 +944,9 @@ static int process_arg(char *p, char *q)
         if (*inname) {
             report_error(ERR_NONFATAL | ERR_NOFILE | ERR_USAGE,
                          "more than one input file specified");
-        } else
-            strcpy(inname, p);
+        } else {
+            copy_filename(inname, p);
+	}
     }
 
     return advance;
@@ -656,7 +965,6 @@ static void process_respfile(FILE * rfile)
     prevarg[0] = '\0';
 
     while (1) {                 /* Loop to handle all lines in file */
-
         p = buffer;
         while (1) {             /* Loop to handle long lines */
             q = fgets(p, bufsize - (p - buffer), rfile);
@@ -702,7 +1010,7 @@ static void process_respfile(FILE * rfile)
             prevargsize += ARG_BUF_DELTA;
             prevarg = nasm_realloc(prevarg, prevargsize);
         }
-        strcpy(prevarg, p);
+        strncpy(prevarg, p, prevargsize);
     }
 }
 
@@ -734,6 +1042,20 @@ static void process_args(char *args)
         process_arg(arg, NULL);
 }
 
+static void process_response_file(const char *file)
+{
+    char str[2048];
+    FILE *f = fopen(file, "r");
+    if (!f) {
+	perror(file);
+	exit(-1);
+    }
+    while (fgets(str, sizeof str, f)) {
+	process_args(str);
+    }
+    fclose(f);
+}
+
 static void parse_cmdline(int argc, char **argv)
 {
     FILE *rfile;
@@ -756,7 +1078,7 @@ static void parse_cmdline(int argc, char **argv)
      * Now process the actual command line.
      */
     while (--argc) {
-        int i;
+	bool advance;
         argv++;
         if (argv[0][0] == '@') {
             /* We have a response file, so process this as a set of
@@ -765,24 +1087,12 @@ static void parse_cmdline(int argc, char **argv)
              * different to the -@resp file processing below for regular
              * NASM.
              */
-            char *str = malloc(2048);
-            FILE *f = fopen(&argv[0][1], "r");
-            if (!str) {
-                printf("out of memory");
-                exit(-1);
-            }
-            if (f) {
-                while (fgets(str, 2048, f)) {
-                    process_args(str);
-                }
-                fclose(f);
-            }
-            free(str);
+	    process_response_file(argv[0]+1);
             argc--;
             argv++;
         }
         if (!stopoptions && argv[0][0] == '-' && argv[0][1] == '@') {
-	    p = get_param(argv[0], argc > 1 ? argv[1] : NULL, &i);
+	    p = get_param(argv[0], argc > 1 ? argv[1] : NULL, &advance);
             if (p) {
 		rfile = fopen(p, "r");
                 if (rfile) {
@@ -793,18 +1103,19 @@ static void parse_cmdline(int argc, char **argv)
                                  "unable to open response file `%s'", p);
             }
         } else
-            i = process_arg(argv[0], argc > 1 ? argv[1] : NULL);
-        argv += i, argc -= i;
+            advance = process_arg(argv[0], argc > 1 ? argv[1] : NULL);
+        argv += advance, argc -= advance;
     }
 
-    if (!*inname)
-        report_error(ERR_NONFATAL | ERR_NOFILE | ERR_USAGE,
-                     "no input file specified");
-
     /* Look for basic command line typos.  This definitely doesn't
        catch all errors, but it might help cases of fumbled fingers. */
-    if (!strcmp(inname, errname) || !strcmp(inname, outname) ||
-	!strcmp(inname, listname))
+    if (!*inname)
+        report_error(ERR_NONFATAL | ERR_NOFILE | ERR_USAGE,
+                     "no input file specified");
+    else if (!strcmp(inname, errname) ||
+	     !strcmp(inname, outname) ||
+	     !strcmp(inname, listname) ||
+	     (depend_file && !strcmp(inname, depend_file)))
 	report_error(ERR_FATAL | ERR_NOFILE | ERR_USAGE,
 		     "file `%s' is both input and output file",
 		     inname);
@@ -831,17 +1142,17 @@ static const char *directives[] = {
 };
 static enum directives getkw(char **directive, char **value);
 
-static void assemble_file(char *fname)
+static void assemble_file(char *fname, StrList **depend_ptr)
 {
     char *directive, *value, *p, *q, *special, *line, debugid[80];
     insn output_ins;
     int i, validid;
     bool rn_error;
-    int32_t seg, offs;
+    int32_t seg;
+    int64_t offs;
     struct tokenval tokval;
     expr *e;
-    int pass, pass_max;
-    int pass_cnt = 0;           /* count actual passes */
+    int pass_max;
 
     if (cmd_sb == 32 && cmd_cpu < IF_386)
         report_error(ERR_FATAL, "command line: "
@@ -849,15 +1160,15 @@ static void assemble_file(char *fname)
 
     pass_max = (optimizing > 0 ? optimizing : 0) + 2;   /* passes 1, optimizing, then 2 */
     pass0 = !(optimizing > 0);  /* start at 1 if not optimizing */
-    for (pass = 1; pass <= pass_max && pass0 <= 2; pass++) {
+    for (passn = 1; pass0 <= 2; passn++) {
         int pass1, pass2;
         ldfunc def_label;
 
-        pass1 = pass < pass_max ? 1 : 2;        /* seq is 1, 1, 1,..., 1, 2 */
-        pass2 = pass > 1 ? 2 : 1;       /* seq is 1, 2, 2,..., 2, 2 */
-        /*      pass0                            seq is 0, 0, 0,..., 1, 2 */
+        pass1 = pass0 == 2 ? 2 : 1;	/* 1, 1, 1, ..., 1, 2 */
+        pass2 = passn > 1  ? 2 : 1;     /* 1, 2, 2, ..., 2, 2 */
+        /* pass0                           0, 0, 0, ..., 1, 2 */
 
-        def_label = pass > 1 ? redefine_label : define_label;
+        def_label = passn > 1 ? redefine_label : define_label;
 
         globalbits = sb = cmd_sb;   /* set 'bits' to command line default */
         cpu = cmd_cpu;
@@ -869,15 +1180,17 @@ static void assemble_file(char *fname)
         global_offset_changed = false;  /* set by redefine_label */
         location.segment = ofmt->section(NULL, pass2, &sb);
         globalbits = sb;
-        if (pass > 1) {
+        if (passn > 1) {
             saa_rewind(forwrefs);
             forwref = saa_rstruct(forwrefs);
             raa_free(offsets);
             offsets = raa_init();
         }
-        preproc->reset(fname, pass1, report_error, evaluate, &nasmlist);
+        preproc->reset(fname, pass1, report_error, evaluate, &nasmlist,
+		       pass1 == 2 ? depend_ptr : NULL);
+	    
         globallineno = 0;
-        if (pass == 1)
+        if (passn == 1)
             location.known = true;
         location.offset = offs = GET_CURR_OFFS;
 
@@ -916,7 +1229,7 @@ static void assemble_file(char *fname)
                             *q++ = '\0';
                             ofmt->symdef(value, 0L, 0L, 3, q);
                         }
-                    } else if (pass == 1) {     /* pass == 1 */
+                    } else if (passn == 1) {
                         q = value;
                         validid = true;
                         if (!isidstart(*q))
@@ -1056,7 +1369,7 @@ static void assemble_file(char *fname)
                             abs_seg = reloc_seg(e);
                             abs_offset = reloc_value(e);
                         }
-                    } else if (pass == 1)
+                    } else if (passn == 1)
                         abs_offset = 0x100;     /* don't go near zero in case of / */
                     else
                         report_error(ERR_PANIC, "invalid ABSOLUTE address "
@@ -1077,13 +1390,13 @@ static void assemble_file(char *fname)
                     }
                     *q++ = 0;
                     if (!validid) {
-                        report_error(pass == 1 ? ERR_NONFATAL : ERR_PANIC,
+                        report_error(passn == 1 ? ERR_NONFATAL : ERR_PANIC,
                                      "identifier expected after DEBUG");
                         break;
                     }
                     while (*p && isspace(*p))
                         p++;
-                    if (pass == pass_max)
+                    if (pass0 == 2)
                         ofmt->current_dfmt->debug_directive(debugid, p);
                     break;
                 case D_WARNING:		/* [WARNING {+|-}warn-name] */
@@ -1167,7 +1480,7 @@ static void assemble_file(char *fname)
                 parse_line(pass1, line, &output_ins,
                            report_error, evaluate, def_label);
 
-                if (!(optimizing > 0) && pass == 2) {
+                if (!(optimizing > 0) && pass0 == 2) {
                     if (forwref != NULL && globallineno == forwref->lineno) {
                         output_ins.forw_ref = true;
                         do {
@@ -1181,7 +1494,7 @@ static void assemble_file(char *fname)
                 }
 
                 if (!(optimizing > 0) && output_ins.forw_ref) {
-                    if (pass == 1) {
+                    if (passn == 1) {
                         for (i = 0; i < output_ins.operands; i++) {
                             if (output_ins.oprs[i].
                                 opflags & OPFLAG_FORWARD) {
@@ -1192,7 +1505,7 @@ static void assemble_file(char *fname)
                                 fwinf->operand = i;
                             }
                         }
-                    } else {    /* pass == 2 */
+                    } else {    /* passn > 1 */
                         /*
                          * Hack to prevent phase error in the code
                          *   rol ax,x
@@ -1220,7 +1533,7 @@ static void assemble_file(char *fname)
 			    output_ins.oprs[1].type &= ~REG_SMASK;
                         }
 
-                    }           /* pass == 2 */
+                    }
 
                 }
 
@@ -1273,7 +1586,7 @@ static void assemble_file(char *fname)
                                 report_error(ERR_NONFATAL,
                                              "bad syntax for EQU");
                         }
-                    } else {    /* pass == 2 */
+                    } else {
                         /*
                          * Special `..' EQUs get processed here, except
                          * `..@' macro processor EQUs which are done above.
@@ -1308,12 +1621,12 @@ static void assemble_file(char *fname)
                                 report_error(ERR_NONFATAL,
                                              "bad syntax for EQU");
                         }
-                    }           /* pass == 2 */
+                    }
                 } else {        /* instruction isn't an EQU */
 
                     if (pass1 == 1) {
 
-                        int32_t l = insn_size(location.segment, offs, sb, cpu,
+                        int64_t l = insn_size(location.segment, offs, sb, cpu,
                                            &output_ins, report_error);
 
                         /* if (using_debug_info)  && output_ins.opcode != -1) */
@@ -1348,6 +1661,16 @@ static void assemble_file(char *fname)
                                     TYS_ELEMENTS(output_ins.oprs[0].
                                                  offset) | TY_TBYTE;
                                 break;
+                            case I_RESO:
+                                typeinfo =
+                                    TYS_ELEMENTS(output_ins.oprs[0].
+                                                 offset) | TY_OWORD;
+                                break;
+                            case I_RESY:
+                                typeinfo =
+                                    TYS_ELEMENTS(output_ins.oprs[0].
+                                                 offset) | TY_YWORD;
+                                break;
                             case I_DB:
                                 typeinfo |= TY_BYTE;
                                 break;
@@ -1369,6 +1692,9 @@ static void assemble_file(char *fname)
 			    case I_DO:
 				typeinfo |= TY_OWORD;
 				break;
+			    case I_DY:
+				typeinfo |= TY_YWORD;
+				break;
                             default:
                                 typeinfo = TY_LABEL;
 
@@ -1386,7 +1712,7 @@ static void assemble_file(char *fname)
                          * flagged as an error on pass 2
                          */
 
-                    } else {    /* pass == 2 */
+                    } else {
                         offs += assemble(location.segment, offs, sb, cpu,
                                          &output_ins, ofmt, report_error,
                                          &nasmlist);
@@ -1414,22 +1740,17 @@ static void assemble_file(char *fname)
                 usage();
             exit(1);
         }
-        pass_cnt++;
-        if (pass > 1 && !global_offset_changed) {
+        if (passn >= pass_max - 2 ||
+	    (passn > 1 && !global_offset_changed))
             pass0++;
-            if (pass0 == 2)
-                pass = pass_max - 1;
-        } else if (!(optimizing > 0))
-            pass0++;
-
-    }                           /* for (pass=1; pass<=2; pass++) */
+    }
 
     preproc->cleanup(0);
     nasmlist.cleanup();
 #if 1
     if (optimizing > 0 && opt_verbose_info)     /*  -On and -Ov switches */
         fprintf(stdout,
-                "info:: assembly required 1+%d+1 passes\n", pass_cnt - 2);
+                "info:: assembly required 1+%d+1 passes\n", passn-3);
 #endif
 }                               /* exit from assemble_file (...) */
 
@@ -1561,18 +1882,16 @@ static void report_error_vc(int severity, const char *fmt, ...)
  * @param severity the severity of the warning or error
  * @return true if we should abort error/warning printing
  */
-static int is_suppressed_warning(int severity)
+static bool is_suppressed_warning(int severity)
 {
     /*
      * See if it's a suppressed warning.
      */
-    return ((severity & ERR_MASK) == ERR_WARNING &&
-            (severity & ERR_WARN_MASK) != 0 &&
-            suppressed[(severity & ERR_WARN_MASK) >> ERR_WARN_SHR]) ||
-        /*
-         * See if it's a pass-one only warning and we're not in pass one.
-         */
-        ((severity & ERR_PASS1) && pass0 == 2);
+    return (severity & ERR_MASK) == ERR_WARNING &&
+	(((severity & ERR_WARN_MASK) != 0 &&
+	  suppressed[(severity & ERR_WARN_MASK) >> ERR_WARN_SHR]) ||
+	 /* See if it's a pass-one only warning and we're not in pass one. */
+	 ((severity & ERR_PASS1) && pass0 != 1));
 }
 
 /**
@@ -1607,18 +1926,20 @@ static void report_error_common(int severity, const char *fmt,
     }
 
     vfprintf(error_file, fmt, args);
-    fputc('\n', error_file);
+    putc('\n', error_file);
 
     if (severity & ERR_USAGE)
         want_usage = true;
 
     switch (severity & ERR_MASK) {
-    case ERR_WARNING:
     case ERR_DEBUG:
         /* no further action, by definition */
         break;
+    case ERR_WARNING:
+	if (!suppressed[0])	/* Treat warnings as errors */
+	    terminate_after_phase = true;
+	break;
     case ERR_NONFATAL:
-        /* hack enables listing(!) on errors */
         terminate_after_phase = true;
         break;
     case ERR_FATAL:
@@ -1656,7 +1977,7 @@ static ListGen *no_pp_list;
 static int32_t no_pp_lineinc;
 
 static void no_pp_reset(char *file, int pass, efunc error, evalfunc eval,
-                        ListGen * listgen)
+                        ListGen * listgen, StrList **deplist)
 {
     src_set_fname(nasm_strdup(file));
     src_set_linnum(0);
@@ -1669,6 +1990,13 @@ static void no_pp_reset(char *file, int pass, efunc error, evalfunc eval,
     no_pp_list = listgen;
     (void)pass;                 /* placate compilers */
     (void)eval;                 /* placate compilers */
+
+    if (deplist) {
+	StrList *sl = nasm_malloc(strlen(file)+1+sizeof sl->next);
+	sl->next = NULL;
+	strcpy(sl->str, file);
+	*deplist = sl;
+    }
 }
 
 static char *no_pp_getline(void)
